#!/usr/bin/perl -w

# Author: Melanie Bahlo

# program to calculate chi-squared test based haplotype sharing statistic to test
# for excessive sharing of haplotypes in cases versus controls.
# This is for cvs to insert version thing into perl script!
# Here is what people have been up to
# $Log: haplo_clusters.pl,v $
# Revision 1.27  2005/03/22 03:47:50  bahlo
# Implemented data imputation in conjunction with permutation testing to
# "sum" over possible data imputations. Also fixed first marker bug. Tested
# this on lots of simulated data which yielded good results (up to 20%
# missing data). Simulated data is in the dir mysim with results.
#
# Revision 1.24  2004/08/19 07:01:35  bahlo
# Minor bug for the -sd qualifier. Wasn't actually adjusting before
# according to input parameter. Now works and allows -sd 0 for multiple
# testing adjustments.
#
# Revision 1.23  2004/08/16 03:54:49  bahlo
#
# Major update of program. -ah 2 now fully functional,
# implemented with ORs and chi-square tests. haplo_sorts
# rewritten more efficiently in hash structure and bug in
# miss 1 option fixed.
#
# Revision 1.20  2003/09/17 07:38:29  bahlo
# Fixed haplotype output fot make it easier to read.
#
# Revision 1.18  2003/09/03 04:57:21  bahlo
# Minor fix to cater for different styles of map.dist files & remove some
# print out re: p-values.
#
# Revision 1.17  2003/07/23 07:02:44  bahlo
# Fixed a minor bug in the output files for the sorts.
#
# Revision 1.16  2003/06/27 01:08:26  bahlo
#
# A fiddly update. Finally got the print out of the sorted haplotypes going
# accurately for multiple clusters. Also fixed the R plot option.
#
# Revision 1.15  2003/06/20 02:08:37  bahlo
# Small modification. To make it perl 5.6 compliant changed all printf %lf's
# to %f's (correct way anyway, but 5.6 cracks down on this).
#
# Revision 1.14  2003/06/20 01:53:54  bahlo
#
# Fixed a few bugs in the output regarding the haplotypes and v still
# sharing. Extensively tested. Sharing is now calculated to >=$share rather
# than >$share. Speedups in Count_freqs and sort_freqs with replacement of
# my own buble sorts with inbuilt PERL sort functions. Approximately 20%
# speedup.
#
# Revision 1.10  2003/05/20 04:24:19  bahlo
# Small update of program. Changed output of files casehaps_sorted and
# controlhaps_sorted so that ID nos of haplotypes are also printed out.
# This is done by storing an additional two vectors: case_ids and
# control_ids.
#
# Revision 1.8  2003/05/14 04:28:17  bahlo
# this is a major rewrite of haplo_clusters.pl
# - help info was fixed.
# - sd fixed
# - -log(p) implemented
# - length of haplotype implemented
# - fixed bug in sort which mucked up sorting with c>1 (and explains why we
# had virtually no difference in results)
# - created "orig" copies of trans, untrans, m, v, m_controls and v_controls
# to facilitate printing out of two further files: casehaps_sorted.out
# and controlhaps_sorted.out which contain the sorted haplotypes for the
# lowest p-value(s). If there is more than one equally low p-value then
# sorted haplotype sets are printed one after the other with a blank line
# separating them.
#
# Revision 1.5  2003/05/12 05:32:35  bahlo
# This is a huge update of haplo_clusters.pl
#
# I have added a number of options.
# -sd for a standard deviation threshhold for the estimator
# recursive calculation of phat and s.d.(phat)
# print out of s.d. and the haplotypes
# s.d. thresholding is currently not working quite right.
#
# Program also prints out permutation at which s.d. threshold was achieved.
#
# I have a few things left I need to do but I wanted to commit this version
# first.
#
# Revision 1.4  2003/03/27 01:58:04  bahlo
#
# new random number generator from CPAN added and map.dist flexibility
# debugged again so that various forms of map.dist can be used.
#
# $Revision: 1.27 $

# ________________________________________________________________


# ---------------Packages--------------------

#use Math::Random::MT qw(srand rand);
#use Math::TrulyRandom;

use Getopt::Long;
use strict;
#use diagnostics; # extended error messages

# --------------global decs--------------------

# -----------for subroutines -----------------
my $time1=time();
my @anchap=(); # the ancestral haplotype on which to carry out the search
my $no=0;
my $cols=2;
my $no_case=0;
my $no_control=0;
my $mark=0;
my $marker=0; # marker counter
my $current_marker=0; # portable $oi from sorter
my @marker_names=(); 
my @trans_haplos=();
my @untrans_haplos=();
my @trans_haplos_orig=();
my @untrans_haplos_orig=();
my @case_ids=(); # case haplotype id nos
my @control_ids=(); # control haplotype id nos
my @haplos=();
my @dists=(); # original map distances
my %new_dists=();
my $logaway; # var for log10
# variables for count_freqs
my $TOT=0; # needed in both count_freqs and count_haplo_freqs
my @unique_alleles=();
my @freq_alls=();
my %freqs=();
my $unique_alls=0;
my @case_unique_alleles=();
my @control_unique_alleles=();
my @case_freq_alls=();
my @control_freq_alls=();
my $test_all;
# variables for count_haplo_freqs
my @unique_haplos=();
my @freq_haplos=();
my $unique_haps=0;
my $haplo_array_size=0;
# variables for sorter
my @order=(); # matrix of size mark*mark to describe the marker order on which to sort the markers
my @v=(); # class members matrix
my @m=(); # haplotype ordering matrix
my @v_controls=(); # class members matrix for controls
my @m_controls=(); # haplotype order matrix for controls
my @v_orig=(); # class members matrix
my @m_orig=(); # haplotype ordering matrix
my @v_controls_orig=(); # class members matrix for controls
my @m_controls_orig=(); # haplotype order matrix for controls
my $classes=0; # class sizes of class members
my @test_marker=(); # stores the permutation no at which sharing stops being calculated
# variables for sharing statistics
my @array_to_sort=(); # array of alleles to be sorted in sort_subarray
my @array_to_sort_controls=(); # array of alleles to be sorted in sort_subarray
my @haplos_to_sort=();
# -----permutation vars--------------
my @new_cases=();
my @new_controls=();
# --------error checks--------------
# if 1 errror check is ON
my $first_check=0; # error check for first marker tests (i.e. ah in {0,1,2})
my $freq_check=0; # error check for frequency determination (allele & haplotype)
my $sorter_check=0; # error check for sorter
my $stat_check=0; # error check for chi-square stats
my $missing_check=0; # error check for missing data fix
my $perm_check=0; # error check for perm tests

# ------------------chi square stats ---------
my @chi_square_perms=(); #permutation chi square test statistics (size: perms*mark)
my @chi_square_orig=(); # original chi square statistics (size: mark)
my @chi_square_orig_parts=(); # original chi square statistics by parts.
my @chi_square_perms_se=(); # standard deviation of phats
my @chi_square_perms_mean=(); # mean of phats
my $chi_square_sum=0; # addition of partial chi-squares.
my @obs=();
my @exp=();
# ----------------------- haplotype storage ---------------
my @store_haplotypes=(); # stores the major haplotypes which are found.
my $orig_sample=""; # flag to indicate that this is the original sample whose 
# chi -square statistic is being calculated, TRUE then. Only then are haplotypes 
# stored for future print out.
# perl will take this as false
my @stop_sharing_marker=(); # vector of markers at which sharing calcs were stopped
my @stop_sharing_marker_orig=(); # vector of markers at which sharing calcs were stopped
# ________________________________________________________________

# the output files...
open(OUT1,">haplo_clusters.out") || print "Can't open file haplo_clusters.out";
open(OUT2,">haplo_clusters_chis.out") || print "Can't open file haplo_clusters_chis.out";

my $opt;
my $miss;
my $perms;
my $c;
my $ah;
my $Rplot;
my $share;
my $Sortfiles;
my $SD;
my $mn;

&GetOptions("opt=i"=>\$opt,"miss=i"=>\$miss,"ah=i",=>\$ah,"perms=i"=>\$perms,"c=i"=>\$c,"share=i"=>\$share,"sd=f"=>\$SD,"mn=s"=>\$mn,"Rplot"=>\$Rplot,"Sortfiles"=>\$Sortfiles);

# to run with the genetic distances use -opt i, if -opt i not used right/left rule is employed.
if(defined($opt)&&defined($perms)){
	if($opt==1){
		print OUT1 "-opt 1 Genetic distances used to decide order of markers to  sort on\n";
		print "-opt 1 Genetic distances used to decide order of markers to sort on\n";
	}
	else{
		print OUT1 "-opt 0 Order of markers to sort on is decided by the right/left rule\n";
		print "-opt 0 Order of markers to sort on is decided by the right/left rule\n";
	}
}
else{
	print "Usage: haplo_clusters.pl -opt i -miss j -perms x [-c y -share u -sd s -mn <file> -ah a]\n";
	print "Author: Melanie Bahlo";
	#$Date: 2005/03/22 03:47:50 $ 
	#$Revision: 1.27 $
	print "\n";
	print "The files casehaps.out and controlhaps.out must be present in the working directory.\n";
	print "Options:\n";
	print "All options not in square brackets must be specified. Others are optionally specified.\n";
	print "-opt i\n";
	print "\t0 use right/left rule to determine marker order\n";
	print "\t1 use genetic distances to determine marker order\n";
	print "\t\tRequires an input file called map.dist with map distances separated by whitespace.\n";
	print "-miss j\n";
	print "\t0 missing data replaced using single marker information\n";
	print "\t1 missing data replaced using 2 marker haplotype information\n";
	print "-perms x\n";
	print "\tx = # of permutations (must be >0)\n";
	print "-ah k\n";
	print "\t0 uses the most frequent case allele as the starting point for the ancestral haplotype (default).\n";
	print "\t1 uses a given ancestral haplotype to start the ancestral haplotype (file ah.in).\n";
	print "\t2 uses the case allele which gives the highest chi-square result to start the ancestral haplotype.\n";
	print "-c y\n";
	print "\ty = # of clusters allowed (default: c = 1)\n";
	print "-share u\n";
	print "\tu = # of case haplotypes at which calculations stop (default: share = 5)\n";
	print "-sd s\n";
	print "\tsd =  standard deviation (p estimator)\n";
	print "-mn <file>\n";
	print "\tAllows the provision of marker names for all output.\n";
	print "-Rplot <file>\n";
	print "\tPrints out a simplified version of haplo_clusters.out for immediate import as a dataframe into R.\n";
	print "-Sortfiles\n";
	print "\tPrints out all haplotypes sorted according to clusters. Produces 2*# of markers extra files.\n";
	print "For more details see the documentation.\n";
	exit(1);
}

if(defined($c)){
	print "-c $c The number of clusters of haplotypes to look for = $c\n";
	print OUT1 "-c $c The number of clusters of haplotypes to look for = $c\n";
}
else{
	$c=1;
	print "The number of clusters of haplotypes to look for = $c (default)\n";
	print OUT1 "The number of clusters of haplotypes to look for = $c (default)\n";
}
if(defined($miss)){
	if($miss==0){
		print "-miss 0 The missing data is replaced randomly using only single marker information.\n";
		print OUT1 "-miss 0 The missing data is replaced randomly using only single marker information.\n";
	}
	else{
		print "-miss 1 The missing data is replaced randomly using the 2 marker haplotype information.\n";
		print OUT1 "-miss 1 The missing data is replaced randomly using the 2 marker haplotype information.\n";
	}
}
else{
	$miss=0;
	print "-miss 0 The missing data is replaced randomly using only single marker information.\n";
	print OUT1 "-miss 0 The missing data is replaced randomly using only single marker information.\n";
}
if(defined($share)){
	print "-share $share The number of haplotypes needed to share = $share\n";
	print OUT1 "-share $share The number of haplotypes needed to share = $share\n";
}
else{
	$share=5;
	print "The number of haplotypes needed to share = $share\n";
	print OUT1 "The number of haplotypes needed to share = $share\n";
}
if(defined($SD)){
	print "-sd $SD The standard deviation to which p values are calculated = $SD*phat\n";
	print OUT1 "-sd $SD The standard deviation to which p values are calculated = $SD*phat\n";
}
else{
	$SD=0.01;
	print "The standard deviation p values are calculated  to 0.01*phat.\n";
	print OUT1 "The standard deviation p values are calculated  to 0.01*phat.\n";
}
if(defined($mn)){
	print "Marker names have been provided and will be used in the output files.\n";
	print OUT1 "Marker names have been provided and will be used in the output files.\n";
}	
$classes=$c;

# options for picking the ancestral haplotype
if(defined($ah)){
	if($ah==0){
		print "-ah $ah Initial ancestral haplotype allele is chosen to be the most frequent case allele.\n";
		print OUT1 "-ah $ah Initial ancestral haplotype allele is chosen to be the most frequent case allele.\n";
	}
	if($ah==1){
		print "-ah $ah Initial ancestral haplotype allele is read in from file.\n";
		print OUT1 "-ah $ah Initial ancestral haplotype allele is read in from file.\n";
		open(IN,"<ah.in") || die print "Can't open file ah.in containing the ancestral haplotype.\n";
		my @temp=<IN>;
		@anchap=split(/\s+/,$temp[0]);
		print "The ancestral haplotype is:\n";
		print OUT1 "The ancestral haplotype is:\n";
		my $ahall;
		foreach $ahall (@anchap){
			$ahall=int($ahall);
			print "$ahall ";
			print OUT1 "$ahall ";
		}
		print "\n";
		print OUT1 "\n";
		undef(@temp);
		undef($ahall);
		close(IN);
		if($c>1){
			print "Currently only one ancestral haplotype can be stipulated.\n";
		}
	}
	if($ah==2){ # currently not implemented
		print "-ah $ah Initial ancestral haplotype allele is chosen to be the case allele";
		print " that gives the highest single marker chi-square test.\n";
		print OUT1 "-ah $ah Initial ancestral haplotype allele is chosen to be the case allele";
		print OUT1 " that gives the highest single marker chi-square test with OR>=1.\n";
	}
}	
else{
	$ah=0;
	print "Initial ancestral haplotype allele is chosen to be the most frequent one.\n";
	print OUT1 "Initial ancestral haplotype allele is chosen to be the most frequent.\n"; 
}	
	
# prints out a special version of haplo_clusters.out only leaving the bare 
# essentials necessary for plotting with the R script plot_sharing.r 
if(defined($Rplot)){
	print "Input necessary for the R script plot_sharing.r has been written into haplo_cluster_R.in\n";
	print OUT1 "Input necessary for the R script plot_sharing.r has been written into haplo_cluster_R.in\n";
}
if(defined($Sortfiles)){
	print "All haplotypes have been sorted according to the ancestral haplotype clusters and are in the files:\n";
	print "casehaps_sorted_*.out and controlhaps_sorted_*.out.\n";
	print OUT1 "All haplotypes have been sorted according to the ancestral haplotype clusters and are in the files:\n";
	print OUT1 "casehaps_sorted_*.out and controlhaps_sorted_*.out.\n";
}
	
# general information:
print "Ordered haplotypes will be printed out to the files casehaps_sorted.out and controlhaps_sorted.out\n";
print "General ouput is in the file haplo_clusters.out\n";
print "Permutation Chi-squares are in haplo_clusters_chis.out (a very large file)\n";
# ________________________________________________________________

# writing casehaps.out data haplos and controlhaps.out data haplotypes into two arrays
# trans_haplos & untrans_haplos
# no of columns that aren't of interest
# I've currently got 2: might make this an option: ID no's and T or U $cols=2; 

# reading in the haplotypes.
get_haplo_input();
#print "Original haplotypes:\n";
#print_orig_haplos();

# finding the order on which to sort the markers
if(defined($mn)){
	read_marker_names();
}

# either read in the marker distances to determine the sort order or
# sort according to the right/left alternate rule
if($opt==1){
    create_marker_order();
}
else{
    create_order_matrix();
}
#print_order();

# checking anchap
if($ah==1){
	if(($#anchap+1)!=$mark){
		print "Ancestral haplotype is not the right length for the given dataset\n";
		exit(1);
	}
}
# --- To the main program ----------------------

main_perm();
close(OUT1);

# --------------------ALL SUBROUTINES ------------------
# --------------------- Miscellaneous subroutines--------------

sub log10{
	return(log($logaway)/log(10));
}

#---------------------initialisation subroutines-------------------------------------

# reading in the haplotypes
sub get_haplo_input{
	my $cnt;
	my $cnt1;
	my @temp;
	my $untrans_mark=0;
	open(IN1,"<casehaps.out") || print "Can't open casehaps.out\n" ;
	open(IN2,"<controlhaps.out") || print "Can't open controlhaps.out\n";
	# get rid of first two columns, id no and T or U
	$cnt=0;
	while(<IN1>){
		@temp=split(/\s+/,$_);
    	$case_ids[$cnt]=shift(@temp); # storing ID numbers of haplotypes
		shift(@temp); # gets rid of T's 
		$cnt1=0;
		foreach(@temp){
    		$trans_haplos[$cnt][$cnt1]=$temp[$cnt1];
    		$trans_haplos_orig[$cnt][$cnt1]=$temp[$cnt1];
			$cnt1+=1;
		}
		$cnt+=1;
	}
	$mark=$#temp+1; # no of markers
	$no_case=$cnt; 
	$cnt=0;
	while(<IN2>){
		@temp=split(/\s+/,$_);
   		$control_ids[$cnt]=shift(@temp); # storing ID numbers of haplotypes
		shift(@temp); # gets rid of T's 
		$cnt1=0;
		foreach(@temp){
    		$untrans_haplos[$cnt][$cnt1]=$temp[$cnt1];
    		$untrans_haplos_orig[$cnt][$cnt1]=$temp[$cnt1];
			$cnt1+=1;
		}
		$cnt+=1;
	}
	$untrans_mark=$#temp+1;
	if($mark!=$untrans_mark){
    	print "The number of markers in the case and control haplotypes is not the same.\n";
    	exit(1);
	}
	$no_control=$cnt;
	close(IN1);
	close(IN2);
	print OUT1 "# of case haplotypes = $no_case\n";
	print OUT1 "# of contol haplotypes = $no_control\n";
	print OUT1 "# of markers = $mark\n";
	print OUT1 "# of perms = $perms\n";
	print "# of case haplotypes = $no_case\n";
	print "# of contol haplotypes = $no_control\n";
	print "# of markers = $mark\n";
	print "# of perms = $perms\n";
}

# ---------------------marker order subroutines-----------------------------------------------

# reads in marker names, these can be in a single line or one on each line.
sub read_marker_names{
	my $i;
	my @arr;
	my @temp1;
	open(IN4,"<$mn") || print "Can't open file $mn\n";
    @arr=<IN4>; # marker order distances
	if($#arr==0){ # map.dist has dists in a single line.
    	@marker_names=split(/\s+/,$arr[0]);
	}
	else{
		foreach(@arr){
			@temp1=split(/\s+/);
			# caters for empty lines behind the last map distance
			if(defined($temp1[0])==0){
				last;
			}
			if($temp1[0] eq ""){ # empty string is first entry in temp1 
				push(@marker_names,$temp1[1]);
			}
			else{# no whitespace
				push(@marker_names,$temp1[0]);
			}
		}
	}	
}

# creates the array of marker order to order on - always goes right first
sub create_order_matrix{
	my $next;
	my $jump;
	my $i;
	my $total_mark=0;
	# total_mark gives the total no of allocated ordered markers
	for($i=0;$i<$mark;$i++){
		$total_mark=0;
		$order[$i][0]=$i;
		for($jump=1;$jump<=$mark;$jump++){
			$next=$i+$jump;
			if($next<$mark&&$total_mark<$mark){
				$total_mark++;
				$order[$i][$total_mark]=$next;
			}
			$next=$i-$jump;
			if($next>=0&&$total_mark<$mark){
				$total_mark++;
				$order[$i][$total_mark]=$next;
			}
		}
	}
}

sub hash_value {$new_dists{$a} <=> $new_dists{$b}}

# creates the array of marker order to order on - now always orders by the next closest marker
# requires an input file marker.order
sub create_marker_order{
    my $i;
    my $j;
	my $key;
    my $going_right=0;
    my $going_left=0;
    my @temp1=();
	my $temp_cnt=0;
	my @arr3;
    my @new_order=();
    open(IN3,"<map.dist") || print "Can't open file map.dist\n";
    @arr3=<IN3>; # marker order distances
	if($#arr3==0){ # map.dist has dists in a single line.
    	@dists=split(/\s+/,$arr3[0]);
	}
	else{
		$temp_cnt=0;
		foreach(@arr3){
			@temp1=split(/\s+/);
			# caters for empty lines behind the last map distance
			if(defined($temp1[0])==0){
				last;
			}
			#print "arr3=$_\n";
			#foreach(@temp1){
			#	print "dist:$_\n";
			#}
			if($temp1[0] eq ""){ # empty string is first entry in temp1 
				if($temp1[1]>0){
					push(@dists,$temp1[1]);
				}
			}
			else{# no whitespace
				if($temp1[0]>0){
					push(@dists,$temp1[0]);
				}
			}
			$temp_cnt+=1;
		}
	}
	#print "dists=@dists\n";
	#exit(1);
    # creating a new distance array matrix of size mark*mark
    # with distance 0.0 for central marker and moving out from there
    for($i=0;$i<$mark;$i++){
		%new_dists=();
		$new_dists{$i}=0.0;
		$going_right=0;
		$going_left=0;
		# new dists is one longer than dists since we have now put in a zero distance as the center
		for($j=$i;$j<($mark-1);$j++){
			$going_right+=$dists[$j];
			$key=$j+1;
			$new_dists{$key}=$going_right;
		}
		for($j=($i-1);$j>=0;$j--){
			$going_left+=$dists[$j];
			$new_dists{$j}=$going_left;
		}
		# the new order
		$order[$i][0]=$i;
		#print "new_dists prior to sort for marker $i\n";
		#foreach $key (keys %new_dists){
		#	print "at marker $key = $new_dists{$key}\n";
		#}
		$j=0;
		for $key (sort hash_value keys %new_dists){
			$order[$i][$j]=$key;
			$j+=1;
		}
   }
}

# --------------------print/error check subroutines -------------------------------------------

# prints order of markers on which to be sorted - an error checking function
sub print_order{
    my $i;
	my $j;
    print("Haplotypes sorted by frequency, \n");
    for($j=0;$j<$mark;$j++){
		print("with sort order determined by the following order, starting at marker ",$j+1,":\n");
		for ($i=0;$i<$mark;$i++) {
	    	print("$order[$j][$i] ");
		}
		print "\n";
    }
    print "\n\n";
}

# printing haplotypes: m vector now determines order
# stop_sharing shows which m vector to print out.
# don't use the later ones as they are not the right ones to use
# use orig m and v as these are the original dataset (not perms!)
sub print_haplos{
    my $i;
    my $j;
	my $m;
	my $l;
	my @print_m_order=();
	my @bound=();
	my $start_search=0;
	my $end_search=0;
	my $start_check=0;
	my $end_check=0;
	my $k;
	my $stop_share;
	my $which_ind;
	$stop_share=$stop_sharing_marker_orig[$marker];
	# get a print order for m.
	# get the positions of v boundaries, store in an k-1 by stop-share sized matrix.
	# creation of my bound (for boundaries) matrix
	for($i=($stop_share-1);$i>0;$i--){
		$bound[$i][0]=$v_orig[$marker][$i][0];
	}
	for($k=1;$k<=$c;$k++){
		for($i=($stop_share-1);$i>0;$i--){
			$bound[$i][$k]=$bound[$i][$k-1]+$v_orig[$marker][$i][$k];
		}
	}
	# special cases: 1) the first cluster from start to finish...
	for($k=0;$k<$c;$k++){
		# special case: final sharing..- no need for sharing checks...
		if($k==0){
			$start_search=0;
			$end_search=$bound[$stop_share-1][0];
		}
		else{
			$start_search=$bound[$stop_share-1][$k-1];
			$end_search=$bound[$stop_share-1][$k];
		}
		for($j=$start_search;$j<$end_search;$j++){	
			$which_ind=$m_orig[$marker][$stop_share-1][$j];
			push @print_m_order,$which_ind;
		}
		# and now the remaining chromosomes with less sharing from cluster k
		for($i=($stop_share-2);$i>0;$i--){
			# start and endpoints for search for haplotype to print
			if($k==0){
				$start_search=0;
				$end_search=$bound[$i][0];
			}
			else{
				$start_search=$bound[$i][$k-1];
				$end_search=$bound[$i][$k];
			}
			HAPLO:for($j=$start_search;$j<$end_search;$j++){
				$which_ind=$m_orig[$marker][$i][$j];
				# start and endpoints to check if haplotype is to be printed.
				if($k==0){
					$start_check=0;
					$end_check=$bound[$i+1][$k];
				}
				else{
					$start_check=$bound[$i+1][$k-1];
					$end_check=$bound[$i+1][$k];
				}
				for($m=$start_check;$m<$end_check;$m++){
					if($m_orig[$marker][$i+1][$m]==$which_ind){ # contained in the previous sort leading
						# to a bigger haplotype.
						next HAPLO;
					}
				}
				push @print_m_order,$which_ind;
			}
		}
	}
	# special case - the non sharing cluster...
	# take all non - sharers from the very first sort..
	$start_search=$bound[1][$c-1];
	$end_search=$no_case;
	for($j=$start_search;$j<$end_search;$j++){
		$which_ind=$m_orig[$marker][1][$j];
		push @print_m_order,$which_ind;
	}
	# printing out all individs
	for($k=0;$k<$no_case;$k++){
		printf OUT3 "%s", "$case_ids[$print_m_order[$k]]";
		print OUT3 " ";
    	for($l=0;$l<$mark;$l++){
			printf OUT3 "%3d", "$trans_haplos_orig[$print_m_order[$k]][$l]";
			print OUT3 " ";
		}
		print OUT3 "\n";
	}
	# the control haplotypes
	@bound=();
	@print_m_order=();	
	for($i=($stop_share-1);$i>0;$i--){
		$bound[$i][0]=$v_controls_orig[$marker][$i][0];
	}
	for($k=1;$k<=$c;$k++){
		for($i=($stop_share-1);$i>0;$i--){
			$bound[$i][$k]=$bound[$i][$k-1]+$v_controls_orig[$marker][$i][$k];
		}
	}
	# special cases: 1) the first cluster from start to finish...
	for($k=0;$k<$c;$k++){
		# special case: final sharing..- no need for sharing checks...
		if($k==0){
			$start_search=0;
			$end_search=$bound[$stop_share-1][0];
		}
		else{
			$start_search=$bound[$stop_share-1][$k-1];
			$end_search=$bound[$stop_share-1][$k];
		}
		for($j=$start_search;$j<$end_search;$j++){	
			$which_ind=$m_controls_orig[$marker][$stop_share-1][$j];
			push @print_m_order,$which_ind;
		}
		# and now the remaining chromosomes with less sharing from cluster k
		for($i=($stop_share-2);$i>0;$i--){
			# start and endpoints for search for haplotype to print
			if($k==0){
				$start_search=0;
				$end_search=$bound[$i][0];
			}
			else{
				$start_search=$bound[$i][$k-1];
				$end_search=$bound[$i][$k];
			}
			HAPLO:for($j=$start_search;$j<$end_search;$j++){
				$which_ind=$m_controls_orig[$marker][$i][$j];
				# start and endpoints to check if haplotype is to be printed.
				if($k==0){
					$start_check=0;
					$end_check=$bound[$i+1][$k];
				}
				else{
					$start_check=$bound[$i+1][$k-1];
					$end_check=$bound[$i+1][$k];
				}
				for($m=$start_check;$m<$end_check;$m++){
					if($m_controls_orig[$marker][$i+1][$m]==$which_ind){ # contained in the previous sort leading
						# to a bigger haplotype.
						next HAPLO;
					}
				}
				push @print_m_order,$which_ind;
			}
		}
	}
	# special case - the non sharing cluster...
	# take all non - sharers from the very first sort..
	$start_search=$bound[1][$c-1];
	$end_search=$no_control;
	for($j=$start_search;$j<$end_search;$j++){
		$which_ind=$m_controls_orig[$marker][1][$j];
		push @print_m_order,$which_ind;
	}
	# printing out all individs
	for($k=0;$k<$no_control;$k++){
		printf OUT4 "%s", "$case_ids[$print_m_order[$k]]";
		print OUT4 " ";
    	for($l=0;$l<$mark;$l++){
			printf OUT4 "%3d", "$untrans_haplos_orig[$print_m_order[$k]][$l]";
			print OUT4 " ";
		}
		print  OUT4 "\n";
	}
}

# printing haplotypes: m vector now determines order
# mainly for error checking: prints all m,v's (not just origs)
# for a given marker $marker after a sort on $current_marker
sub print_haplos_temp{
    my $i;
    my $j;
	# casehaplos first
	print "Sort after marker sort on $current_marker th marker, starting sort at $marker\n";
	print "Case haplotypes\n";
    for($i=0;$i<$no_case;$i++){
		print "$m[$marker][$current_marker][$i] ";
    	for($j=0;$j<$mark;$j++){
			# m[$marker][$l][$m] is the current order of haplos
			# $l is determined by when the sharing calcs were stopped.
			print "$trans_haplos[$m[$marker][$current_marker][$i]][$j]";
			print " ";
		}
		print "\n";
    }
	print "Control haplotypes\n";
	# controlhaplos
    for($i=0;$i<$no_control;$i++){
		print "$m_controls[$marker][$current_marker][$i] ";
     	for($j=0;$j<$mark;$j++){
			print "$untrans_haplos[$m_controls[$marker][$current_marker][$i]][$j]";
			print " ";
		}
		print "\n";
    }
    print "\n";
}

# printing the unsorted original haplotypes
# useful for error checking prior to sorting e.g. for missing data replacement
# as m and v are not defined until later.
# BUT gets overwritten with permutations later on in the piece
# so datasets may be different.
sub print_orig_haplos{
    my $i;
    my $j;
	print "Case haplotypes:\n";
	for($i=0;$i<$no_case;$i++){
   		for($j=0;$j<$mark;$j++){
			print "$trans_haplos[$i][$j] ";
    	}
    	print "\n";
	}
	print "Control haplotypes:\n";
	for($i=0;$i<$no_control;$i++){
   		for($j=0;$j<$mark;$j++){
			print "$untrans_haplos[$i][$j] ";
    	}
    	print "\n";
	}
}

# prints new permuted samples of haplotypes
sub print_new_haplos{
my $i1;
my $j1;
	print "new sample of case haplotypes\n";
	for($i1=0;$i1<$no_case;$i1++){
		for($j1=0;$j1<$mark;$j1++){
			print "$new_cases[$i1][$j1] ";
		}
		print "\n";
	}
	print "new sample of control haplotypes\n";
	for($i1=0;$i1<$no_control;$i1++){
		for($j1=0;$j1<$mark;$j1++){
			print "$new_controls[$i1][$j1] ";
		}
		print "\n";
	}
}

# print allele frequencies
sub print_freq{
    my $i;
    printf "%10s","Allele at Marker ";
    print $marker+1;
    printf "%10s", "Freqs";
    print "\n";
	$TOT=0;
    print "-------------------------\n";
    for($i=0;$i<=$#freq_alls;$i++){
		printf "%15d", "$unique_alleles[$i]";
		printf "%10d", "$freq_alls[$i] \n";
		print "\n";
		$TOT+=$freq_alls[$i];
    }
    print "-------------------------\n";
    printf "%25d", "$TOT";
    print "\n";
}

# print haplotype frequencies
sub print_haplo_freq{
    my $i2;
    printf "%30s","2 Marker Haplotypes at Marker ";
    print $marker+1;
    printf "%10s", "Freqs";
    print "\n";
	$TOT=0;
    print "-------------------------\n";
    for($i2=0;$i2<$unique_haps;$i2++){
		printf "%3d %3d", "$unique_haplos[$i2][0]","$unique_haplos[$i2][1]";
		print "\t\t";
		printf "%10d", "$freq_haplos[$i2] \n";
		print "\n";
		$TOT+=$freq_haplos[$i2];
    }
    print "-------------------------\n";
    printf "%25d", "$TOT";
    print "\n";
}

# calculating and printing the table of p-values calculated from the empirical distribution
sub p_values_print{
	my $i;
	my $j;
	my $k;
	my $l;
	my $total_dist=0;
	my $count_haplos=0;
	my $temp=0;
	my @min_p_markers=();
	my $min_p;
	my $oi;
	my $cnt=0;
	my $cah_sorted_out;
	my $coh_sorted_out;
	my @order1=();
	my @order2=();
	my $tmp;
	if(defined($Rplot)){
		open(OUT3,">haplo_clusters_R.in") || print "Can't open R input file\n";
	}
	if($opt==0){
		if(defined($mn)){
			print OUT1 "Marker\tMarker #\tChi_Square\tp\tsd(p)\t-log(p)\tOKorMAX\tperms\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tMarker #\tChi_Square\tp\tsd(p)\t-log(p)\tOKorMAX\tperms\n";
			}
		}
		else{
			print OUT1 "Marker\tChi_Square\tp\tsd(p)\t-log(p)\tOKorMAX\tperms\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tChi_Square\tp\tsd(p)\t-log(p)\tOKorMAX\tperms\n";
			}
		}
	}
	else{
		if(defined($mn)){
			print OUT1 "Marker\tMapdistance\tChi_Square\tp\tsd(p)\t-log(p)\tperms\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tMapdistance\tChi_Square\tp\tsd(p)\t-log(p)\tOKorMAX\tperms\n";
			}
		}
		else{
			print OUT1 "Marker\tMapdistance\tChi_Square\tp\tsd(p)\t-log(p)\tperms\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tMapdistance\tChi_Square\tp\tsd(p)\t-log(p)\tOKorMAX\tperms\n";
			}
		}
	}
	print OUT1 "====================================================\n";
	$min_p=1.0;
	for($j=0;$j<$mark;$j++){
		if(defined($chi_square_perms_mean[$j])==$min_p){
			push(@min_p_markers,$j);
		}
		if(defined($chi_square_perms_mean[$j])<$min_p){
			$min_p=$chi_square_perms_mean[$j];
			@min_p_markers=(); # re initialize min_p_marker vector.
			$min_p_markers[0]=$j;
		}
	}
	for($j=0;$j<$mark;$j++){
		if(defined($mn)){
			print OUT1 "$marker_names[$j]\t",$j+1;
			if(defined($Rplot)){
				print OUT3 $marker_names[$j];
			}
		}
		else{
			print OUT1 $j+1;
			if(defined($Rplot)){
				print OUT3 $j+1;
			}
		}
		if($opt==1){
			if($j==0){
				$total_dist=0;
			}
			else{
				$total_dist+=$dists[$j-1];
			}
			print OUT1 "\t$total_dist\t\t";
			if(defined($Rplot)){
				print OUT3 "\t$total_dist\t\t";
			}
		}
		else{
			print OUT1 "\t\t";
			if(defined($Rplot)){
				print OUT3 "\t\t";
			}
		}
		printf OUT1 "%1.2f",$chi_square_orig[$j];
		print OUT1 "\t\t";
		printf OUT1 "%1.2e",$chi_square_perms_mean[$j];
		print OUT1 "\t";
		printf OUT1 "%1.2e",$chi_square_perms_se[$j];
		print OUT1 "\t";
		if(defined($Rplot)){
			printf OUT3 "%1.2f",$chi_square_orig[$j];
			print OUT3 "\t\t";
			printf OUT3 "%1.2e",$chi_square_perms_mean[$j];
			print OUT3 "\t";
			printf OUT3 "%1.2e",$chi_square_perms_se[$j];
			print OUT3 "\t";
		}
		if($chi_square_perms_mean[$j]>0){
			$logaway=$chi_square_perms_mean[$j];
			$temp=log10();
			$temp*= -1;
			printf OUT1 "%1.2f",$temp;
			if(defined($Rplot)){
				printf OUT3 "%1.2f",$temp;
				print OUT3 "\tOK";
			}
		}
		else{
			printf OUT1 ">";
			$logaway=$perms;
			$temp=log10();
			printf OUT1 "%1.2f",$temp;
			# make this approximation for the R plot so that you can plot it.
			if(defined($Rplot)){
				printf OUT3 "%1.2f",$temp;
				print OUT3 "\tMAX";
			}
		}
		if($test_marker[$j] ne ""){
			print OUT1 "\t$test_marker[$j]\n";
			if(defined($Rplot)){
				print OUT3 "\t$test_marker[$j]\n";
			}
		}
		else{
			print OUT1 "\t$perms\n";
			if(defined($Rplot)){
				print OUT3 "\t$perms\n";
			}
		}
	}
	print OUT1 "===================================================\n";
	# print out the sorted haplotypes 
	if(defined($Sortfiles)){
		#print "min_p = $min_p, min_p_markers=@min_p_markers\n";
		for($i=0;$i<=$#min_p_markers;$i++){
			if(defined($mn)){
				$cah_sorted_out="casehaps_sorted_".$marker_names[$min_p_markers[$i]].".out";
				$coh_sorted_out="controlhaps_sorted_".$marker_names[$min_p_markers[$i]].".out";
			}
			else{
				$cah_sorted_out="casehaps_sorted_".$min_p_markers[$i].".out";
				$coh_sorted_out="controlhaps_sorted_".$min_p_markers[$i].".out";
			}
			open(OUT3,">$cah_sorted_out") || print "Can't open file $cah_sorted_out";
			open(OUT4,">$coh_sorted_out") || print "Can't open file $coh_sorted_out";
			#print "Marker $marker_names[$min_p_markers[$i]] had the smallest p-value = $min_p.\n";
			#print "Sorts are printed out in the files $cah_sorted_out and $coh_sorted_out.\n";
			$marker=$min_p_markers[$i];
			print_haplos();
			close(OUT3);
			close(OUT4);
		}
	}
	# routines to print out the shared haplotypes
	print OUT1 "Table of haplotypes\n";
	print OUT1 "Marker\tCluster\tHaplotype\tLength(Haplotype)\n";
	print OUT1 "\t\t";
	print OUT1 "\n===================================================\n";
	for($j=0;$j<$mark;$j++){
		$cnt=0;
		CLUSTER: for($k=0;$k<$c;$k++){
			print OUT1 "Cluster ",$k+1,"\n";
			print OUT1 "\t\t\t";
			if(defined($mn)){
				for($l=0;$l<$mark;$l++){
					if(defined($store_haplotypes[$j][$l][$k])){
						print OUT1 "$marker_names[$l]\t";
					}
				}
				print OUT1 "\n";
			}
			print OUT1 "\t\t\t";
			for($l=0;$l<$mark;$l++){
				if(defined($store_haplotypes[$j][$l][$k])){
					print OUT1 $l+1,"\t";
				}
			}
			print OUT1 "\n";
			$count_haplos=0;			
			if(defined($mn)){
				print OUT1 "$marker_names[$j]\t",$j+1,"\t";
			}
			else{
				print OUT1 $j+1,"\t";
			}
			for($l=0;$l<$mark;$l++){
				if(defined($store_haplotypes[$j][$l][$k])){
					print OUT1 "\t$store_haplotypes[$j][$l][$k]";
				}
				#print "store_haplotypes[$j][$l][$k]=$store_haplotypes[$j][$l][$k]\n";
				$count_haplos+=1;
			}
			#print OUT1 "\t$count_haplos";
			print OUT1 "\n# of haplos:";
			$cnt=1;
			print OUT1 "\t"; # formatting to bring in line with last line
			for($l=0;$l<$mark;$l++){
				$order1[$l]=$order[$j][$l]; # the sort order is not the same as the v order!
			}
			for($l=0;$l<$mark;$l++){
				$order2[$order1[$l]]=$cnt;
				$cnt+=1;
			}
			for($l=0;$l<$mark;$l++){
				$tmp=$order2[$l];				
				if(defined($store_haplotypes[$j][$l][$k])){
				#if(defined($v_orig[$j][$tmp][$k])){
					print OUT1 "\t$v_orig[$j][$tmp][$k]";
				}
			}
			print OUT1 "\nChi-square:";
			print OUT1 "\t";# formatting to bring in line with last line
			for($l=0;$l<$mark;$l++){
				$oi=$order2[$l]-1;
				if(defined($store_haplotypes[$j][$l][$k])){
				#if(defined($chi_square_orig_parts[$j][$oi][$k])){
					print OUT1 "\t";
					printf OUT1 "%2.2f",$chi_square_orig_parts[$j][$oi][$k];
				}
			}					
			print OUT1 "\n\n";
		}
	}
	print OUT1 "===================================================\n";
	close(OUT3);
}

# calculating and printing the table of results when no permutation testing is possible (perms<2)
sub results_print{
	my $i;
	my $j;
	my $k;
	my $l;
	my $total_dist=0;
	my $count_haplos=0;
	my $temp=0;
	my @min_p_markers=();
	my $min_p;
	my $oi;
	my $cnt=0;
	my $cah_sorted_out;
	my $coh_sorted_out;
	my @order1=();
	my @order2=();
	my $tmp;
	if(defined($Rplot)){
		open(OUT3,">haplo_clusters_R.in") || print "Can't open R output file\n";
	}
	if($opt==0){
		if(defined($mn)){
			print OUT1 "Marker\tMarker #\tChi_Square\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tMarker #\tChi_Square\n";
			}
		}
		else{
			print OUT1 "Marker\tChi_Square\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tChi_Square\n";
			}
		}
	}
	else{
		if(defined($mn)){
			print OUT1 "Marker\tMapdistance\tChi_Square\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tMapdistance\tChi_Square\n";
			}
		}
		else{
			print OUT1 "Marker\tMapdistance\tChi_Square\n";
			if(defined($Rplot)){
				print OUT3 "Marker\tMapdistance\tChi_Square\n";
			}
		}
	}
	print OUT1 "====================================================\n";
	for($j=0;$j<$mark;$j++){
		if(defined($mn)){
			print OUT1 "$marker_names[$j]\t",$j+1;
			if(defined($Rplot)){
				print OUT3 $marker_names[$j];
			}
		}
		else{
			print OUT1 $j+1;
			if(defined($Rplot)){
				print OUT3 $j+1;
			}
		}
		if($opt==1){
			if($j==0){
				$total_dist=0;
			}
			else{
				$total_dist+=$dists[$j-1];
			}
			print OUT1 "\t$total_dist\t\t";
			if(defined($Rplot)){
				print OUT3 "\t$total_dist\t\t";
			}
		}
		else{
			print OUT1 "\t\t";
			if(defined($Rplot)){
				print OUT3 "\t\t";
			}
		}
		printf OUT1 "%1.2f",$chi_square_orig[$j];
		print OUT1 "\n";
		if(defined($Rplot)){
			printf OUT3 "%1.2f",$chi_square_orig[$j];
			print OUT3 "\n";
		}
	}
	print OUT1 "===================================================\n";
	# print out the sorted haplotypes 
	if(defined($Sortfiles)){
		#print "min_p = $min_p, min_p_markers=@min_p_markers\n";
		for($i=0;$i<=$#min_p_markers;$i++){
			if(defined($mn)){
				$cah_sorted_out="casehaps_sorted_".$marker_names[$min_p_markers[$i]].".out";
				$coh_sorted_out="controlhaps_sorted_".$marker_names[$min_p_markers[$i]].".out";
			}
			else{
				$cah_sorted_out="casehaps_sorted_".$min_p_markers[$i].".out";
				$coh_sorted_out="controlhaps_sorted_".$min_p_markers[$i].".out";
			}
			open(OUT3,">$cah_sorted_out") || print "Can't open file $cah_sorted_out";
			open(OUT4,">$coh_sorted_out") || print "Can't open file $coh_sorted_out";
			#print "Marker $marker_names[$min_p_markers[$i]] had the smallest p-value = $min_p.\n";
			#print "Sorts are printed out in the files $cah_sorted_out and $coh_sorted_out.\n";
			$marker=$min_p_markers[$i];
			print_haplos();
			close(OUT3);
			close(OUT4);
		}
	}
	# routines to print out the shared haplotypes
	print OUT1 "Table of haplotypes\n";
	print OUT1 "Marker\tCluster\tHaplotype\tLength(Haplotype)\n";
	print OUT1 "\t\t";
	print OUT1 "\n===================================================\n";
	for($j=0;$j<$mark;$j++){
		$cnt=0;
		CLUSTER: for($k=0;$k<$c;$k++){
			print OUT1 "Cluster ",$k+1,"\n";
			print OUT1 "\t\t\t";
			if(defined($mn)){
				for($l=0;$l<$mark;$l++){
					if(defined($store_haplotypes[$j][$l][$k])){
						print OUT1 "$marker_names[$l]\t";
					}
				}
				print OUT1 "\n";
			}
			print OUT1 "\t\t\t";
			for($l=0;$l<$mark;$l++){
				if(defined($store_haplotypes[$j][$l][$k])){
					print OUT1 $l+1,"\t";
				}
			}
			print OUT1 "\n";
			$count_haplos=0;			
			if(defined($mn)){
				print OUT1 "$marker_names[$j]\t",$j+1,"\t";
			}
			else{
				print OUT1 $j+1,"\t";
			}
			for($l=0;$l<$mark;$l++){
				if(defined($store_haplotypes[$j][$l][$k])){
					print OUT1 "\t$store_haplotypes[$j][$l][$k]";
				}
				#print "store_haplotypes[$j][$l][$k]=$store_haplotypes[$j][$l][$k]\n";
				$count_haplos+=1;
			}
			#print OUT1 "\t$count_haplos";
			print OUT1 "\n# of haplos:";
			$cnt=1;
			print OUT1 "\t"; # formatting to bring in line with last line
			for($l=0;$l<$mark;$l++){
				$order1[$l]=$order[$j][$l]; # the sort order is not the same as the v order!
			}
			for($l=0;$l<$mark;$l++){
				$order2[$order1[$l]]=$cnt;
				$cnt+=1;
			}
			for($l=0;$l<$mark;$l++){
				$tmp=$order2[$l];				
				if(defined($store_haplotypes[$j][$l][$k])){
				#if(defined($v_orig[$j][$tmp][$k])){
					print OUT1 "\t$v_orig[$j][$tmp][$k]";
				}
			}
			print OUT1 "\nChi-square:";
			print OUT1 "\t";# formatting to bring in line with last line
			for($l=0;$l<$mark;$l++){
				$oi=$order2[$l]-1;
				if(defined($store_haplotypes[$j][$l][$k])){
				#if(defined($chi_square_orig_parts[$j][$oi][$k])){
					print OUT1 "\t";
					printf OUT1 "%2.2f",$chi_square_orig_parts[$j][$oi][$k];
				}
			}					
			print OUT1 "\n\n";
		}
	}
	print OUT1 "===================================================\n";
	close(OUT3);
}

# --------------------sort subroutines----------------------------------------------

# gets frequencies for the array _to_sort
sub count_freqs{
	my @sorted_array;
	my $key;
	%freqs=();
	#@sorted_array=sort {$a<=>$b} @array_to_sort;
	# creating the freqs hash or contingency table
	foreach $key (@array_to_sort){
		$freqs{$key}+=1;
	}
	# remove the 0's, delete is a hash specific function.
	delete $freqs{0};
	$unique_alls=scalar(keys %freqs);
}

# formerly part of sort freqs..
# sorting the alleles by allele frequency
sub sort_freqs{
my $key;
my $i=0;
	@freq_alls=();
	@unique_alleles=();
	sub hash_by_value{$freqs{$b}<=>$freqs{$a}}
	for $key (sort hash_by_value keys %freqs){
		$freq_alls[$i]=$freqs{$key};
		$unique_alleles[$i]=$key;
		$i+=1;
	}
	if($freq_check){
		print_freq();
	} 
}

# sorting the alleles by the ancestral haplotype
# currently restricted to c=1
sub sort_freqs_by_anchap{
my $key;
my $i=0;
	# first entry in freq_alls is the given ancestral haplotype allele
	if(defined($freqs{$anchap[$current_marker]})){
		$freq_alls[0]=$freqs{$anchap[$current_marker]};
	}
	else{
		$freq_alls[0]=0;
		#$freqs{$anchap[$current_marker]}=0;
	}
	$i=1;
	$unique_alleles[0]=$anchap[$current_marker];
	# sort the rest according to value
	for $key (sort hash_by_value keys %freqs){
		if($key ne $anchap[$current_marker]){
			$freq_alls[$i]=$freqs{$key};
			$unique_alleles[$i]=$key;
			$i+=1;
		}
	}
	$unique_alls=$#unique_alleles+1;
	if($first_check){
		print "sort_freqs_by_anchap after sort on $current_marker\n";
		print_freq();
	} 
}

# test statistic will be the binomial test statistic here rather than
# chi-square because it gives us the DIRECTION of the test as we want 
# susceptibility genes, not protective genes ie. O(X)>E(X) where
# X is the number of observed copies of allele X in the cases
# stat^2 = chi-square 1 d.f.
sub chi_square_first{
my $j;
my $case_sum=0;
my $control_sum=0;
my $control_test_all;
my $OR;
my $test_statistic;
	# splice out the not wanted allele counts
	for($j=0;$j<=$#case_freq_alls;$j++){
		$case_sum+=$case_freq_alls[$j];
	}
	# find prob(of allele $test-all) in controls:
	$control_test_all=-1;
	for($j=0;$j<=$#control_freq_alls;$j++){
		if($control_unique_alleles[$j]==$case_unique_alleles[$test_all]){
			$control_test_all=$j;
		}
		$control_sum+=$control_freq_alls[$j];
	}
	if($control_test_all==-1 && $first_check){
		print "No test possible for allele $case_unique_alleles[$test_all] because there are no controls with this allele\n";
	}
	# OR test: OR=ad/bc
	if((($control_test_all>-1)&&$case_sum>$case_freq_alls[$test_all])&& $ control_sum>$control_freq_alls[$control_test_all]){
		$OR=$case_freq_alls[$test_all]/$control_freq_alls[$control_test_all];
		$OR/=($case_sum-$case_freq_alls[$test_all])/($control_sum-$control_freq_alls[$control_test_all]);
	}
	else{
		$OR=0; # this isn't true actually because the OR is actually infinity but this is the way
		# I'm coping with it.
	}
	# printing the chi-square table:
	if($first_check&& $OR>=0){
		print "\t\t\tCases\tControls\n";
		if($control_test_all>-1){
			print "Allele $case_unique_alleles[$test_all]\t$case_freq_alls[$test_all]\t$control_freq_alls[$control_test_all]\n";
		}
		else{
			print "Allele $case_unique_alleles[$test_all]\t$case_freq_alls[$test_all]\t0\n";
		}
		print "Not Allele $case_unique_alleles[$test_all]\t";
		print $case_sum-$case_freq_alls[$test_all],"\t";
		if($control_test_all > -1){
			print $control_sum-$control_freq_alls[$control_test_all];
		}
		else{
			print "$control_sum";
		}
		print "\n";
		print "OR = $OR\n";
	}
	# probability of observing a case with allele x under H_o
	# just use chi_square_stat, but only if OR is undefined (that's actually infinity
	# when case_freq=0 or >0
	# what format for data does chi-square stat want?
	@obs=();
	$obs[0][0]=$case_freq_alls[$test_all];
	if($control_test_all> -1){
		$obs[0][1]=$control_freq_alls[$control_test_all];
		$obs[1][1]=$control_sum-$control_freq_alls[$control_test_all];
	}
	else{
		$obs[0][1]=0;
		$obs[1][1]=$control_sum;
	}
	$obs[1][0]=$case_sum-$case_freq_alls[$test_all];	
	if($perm_check){
		print "Anchap 2 chi-square-stat calculation\n";
	}
	if($OR>=1||$OR==0){
		$test_statistic=chi_square_stat();
	}
	else{
		$test_statistic=0;
	}
	if($first_check){
		print "OR=$OR, test_statistic=$test_statistic\n";
	}
	return($test_statistic);
}

# re-arranges frequency table for alleles according to most significant chi-square value
# or single marker test.
# do this by using the chi-square test but only doing the top five of the alleles
# others are going to have too low an allele frequency
sub sort_freqs_by_chi{
my $key;
my $i;
my $j;
my $all_lim;
my $mystat;
my $case_sum=0;
my @remain_order=();
my @top_order=(); # allele top order but by position in original array
my @chi_top_order=(); # binomial stat top order
my @chi_remain_order=();
	# start by doing usual sort
	# special case of low polymorphic markers (e.g.SNPs)
	# What happens with SNPs??? What do we do?
	# finding the chi-squares for the top ones
	for($i=0;$i<=$#case_unique_alleles;$i++){
		$case_sum+=$case_freq_alls[$i];
	}
	if($first_check){
		print "Cases:\n";
		print "Alleles: @case_unique_alleles\n";
		print "Freqs: @case_freq_alls\n";
		print "Controls:\n";
		print "Alleles: @control_unique_alleles\n";
		print "Freqs: @control_freq_alls\n";
	}
	ALLELE:for($i=0;$i<=$#case_unique_alleles;$i++){
		# don't bother testing if case frequencies are lower than this
		# and OR is <1. We need OR >1 for susceptibility haplotype. Especially
		# important for SNP haplotypes
		# add the marker to the top of the pile with highest at the top
		if($case_freq_alls[$i]/$case_sum>0.05){
			if($case_freq_alls[$i]/$case_sum==1){ # OR infinity case...
				splice(@chi_top_order,$j,0,9999.99);
				splice(@top_order,$j,0,$i);
				next ALLELE;
			}	
			$test_all=$i;
			$mystat=chi_square_first($test_all);
			#if($mystat>0){
				for($j=0;$j<=$#top_order;$j++){
					if($mystat>$chi_top_order[$j]){
						# insert into appropriate position
						splice(@chi_top_order,$j,0,$mystat);
						splice(@top_order,$j,0,$i);
						next ALLELE;
					}
				}
				# add at the end since it this is currently the smallest
				push(@chi_top_order,$mystat);
				push(@top_order,$i);
			#}
		}
		# all the other alleles go down the bottom in an unsorted order
		else{
			push(@remain_order,$i);
			push(@chi_remain_order,0);
		}
	}
	push(@top_order,@remain_order);
	push(@chi_top_order,@chi_remain_order);
	if($first_check){
		print "The order of alleles according to the chi-square statistic\n";
		print "Allele\tchi-square\n";
		print "-----------------\n";
		for($i=0;$i<=$#top_order;$i++){
			print "$case_unique_alleles[$top_order[$i]]\t$chi_top_order[$i]\n";
		}
		print "-------------------\n";
		print "\n";
	}
	# resort the case frequencies according to this new ordering
	$i=0;
	@freq_alls=();
	@unique_alleles=();
	for $key (@top_order){
		$freq_alls[$i]=$case_freq_alls[$key];
		$unique_alleles[$i]=$case_unique_alleles[$key];
		$i+=1;
	}
	$unique_alls=$#freq_alls+1;
	if($first_check){
		print "Sort after sort on first marker according to 2 by 2 chi-squares (ah ==2)\n";
		print_freq();
	}
}

# gets frequencies for the array array_to_sort,unlike the last function we
# are now sorting haplotypes made up of pairs of markers, not a single marker sort
# want to derive haplotype allele frequencies.
sub count_haplo_freqs{
	my $i;
	my $key;
	my $TOT;
	my @hap=();
	$unique_haps=0;
	@unique_haplos=();
	@freq_haplos=();
	%freqs=();
	# getting the frequencies and vector of different allele labels
	for ($i=0; $i<$haplo_array_size;$i++) {
		if($haplos_to_sort[$i][0]!=0 && $haplos_to_sort[$i][1]!=0){
			$key=$haplos_to_sort[$i][0]."/".$haplos_to_sort[$i][1];
			$freqs{$key}+=1;
		}
	}
	#foreach $key (keys %freqs){
	#	print "haplotype $key has freq $freqs{$key}\n";
	#}
	# sort haplotype freqs from highest to lowest
	$i=0;
	if($missing_check){
		print "Haplotype\tFrequency\n";
	}
	sub haps_by_value{$freqs{$b}<=>$freqs{$a}};
	if($missing_check){
		print "-----------------------------------\n";
	}
	for $key (sort haps_by_value keys %freqs){
		$freq_haplos[$i]=$freqs{$key};
		@hap=split(/\//,$key);
		$unique_haplos[$i][0]=$hap[0];
		$unique_haplos[$i][1]=$hap[1];
		if($missing_check){
			print "$unique_haplos[$i][0]/$unique_haplos[$i][1] $freq_haplos[$i]\n";
		}
		$TOT+=$freq_haplos[$i];
		$i+=1;
	}
	$unique_haps=$i;
	if($missing_check){
		print "-----------------------------------\n";
		print "\t\t$TOT\n";
	}
}

# choosing a new allele to replace the missing one randomly from set
sub choose_allele{
	my $i;
	my $sub_total=0;
	my $new_allele;
	my $new;
	$new=int(rand($TOT));
	$sub_total=$freq_alls[0]; #intialise with the largest allele frequency, this is the most
	#likely to be chosen, hence makes this more efficient.
	for($i=0;$i<$unique_alls;$i++){
		if($new>$sub_total){
			$new_allele=$unique_alleles[$i-1];
			return($new_allele);
		}
		$sub_total+=$freq_alls[$i];
	}
	# can only be the last allele
	$new_allele=$unique_alleles[$unique_alls-1];
	return($new_allele);
}

# choosing a haplotype randomly from the set of available haplotypes
sub choose_haplo{
	my $i;
	my $j;
	my $new_allele;
	my $new;
	my $hap;
	my $sub_total=0;
	$TOT=0;
	foreach $hap (@freq_haplos){
		$TOT+=$hap;
	}
	$new=int(rand($TOT));
	#print "new=$new,TOT=$TOT,unique_haps=$unique_haps\n";
	$sub_total=$freq_haplos[0]; #intialise with the largest allele frequency, this is the most
	#likely to be chosen, hence makes this more efficient.
	for($i=0;$i<$unique_haps;$i++){
		if($new<$sub_total){
			$new_allele=$unique_haplos[$i][1]; # the missing allele becomes the second allele of the chosen
			# haplotype
			return($new_allele);
		}
		$sub_total+=$freq_haplos[$i+1];
	}
	# can only be the last haplotype
	$new_allele=$unique_haplos[$unique_haps-1][1];
	return($new_allele);
}

# January 2002 fixed up
# seeks out zero alleles and replaces them with an allele randomly chosen either by method 1 or method 0
# method 0 entails the drawing of a random allele from just the single marker data
# method 1 derives the missing allele from the haplotype distribution of that marker and its left neighbour
# this function runs once only, prior to sorting. Unlike previous incarnations it does not change 
# the substitution according to sort (no bias hopefully!)
# The case and control haplotypes are separately dealt with.
sub fix_missing_data{
	my $i;
	my $j;
	my $k;
	my $l;
	my $j1;
	my $array_size=0;
	my $new_allele;
	# using count_freqs for 1st option ie miss -0
	if($miss==0){ 
		# -----------the case haplotypes-----------------
		for($i=0;$i<$mark;$i++){
			@array_to_sort=();
			for($j=0;$j<$no_case;$j++){
				push @array_to_sort,$trans_haplos[$j][$i];
			}
			$marker=$i; # this is needed in count_freqs for storage purposes
			$array_size=$#array_to_sort+1;
			count_freqs();
			sort_freqs();
			if($missing_check==1){
				print "Before 0 removal (cases) (option: -miss 0):\n";
				print_freq();
			}
			# count freqs generates a variable called $TOT which is the total no of non-zero alleles
			for($j=0;$j<$no_case;$j++){
				if($trans_haplos[$j][$i]==0){
					$new_allele=choose_allele();
					if($missing_check==1){
						print "Check: old allele trans_haplos[$j][$i] = $trans_haplos[$j][$i], new allele = $new_allele\n";
					}
					$trans_haplos[$j][$i]=$new_allele;
				}
			}
		}
		# ---------------the control haplotypes--------------------
		for($i=0;$i<$mark;$i++){
			@array_to_sort=();
			for($j=0;$j<$no_control;$j++){
				push @array_to_sort,$untrans_haplos[$j][$i];
			}
			$array_size=$#array_to_sort+1;
			$marker=$i; # this is needed in count_freqs for storage purposes
			count_freqs();
			sort_freqs();
			if($missing_check==1){
				print "Before 0 removal (controls) (option: -miss 0):\n";
				print_freq();
			}
			for($j=0;$j<$no_control;$j++){
				if($untrans_haplos[$j][$i]==0){
					$new_allele=choose_allele();
					if($missing_check==1){
						print "Check: old allele untrans_haplos[$j][$i] = $untrans_haplos[$j][$i], new allele = $new_allele\n";
					}
					$untrans_haplos[$j][$i]=$new_allele;
				}
			}
		}
	}
	else{
		# using count_haplo_pairs for 2nd option ie miss -1
		# only get haplo freqs for markers 1 to mark-1, use single marker freqs for marker 0
		# ---------------the case haplotypes--------------------------
		# -------------special case of 1st marker
		@array_to_sort=();
		for($j=0;$j<$no_case;$j++){
			push @array_to_sort,$trans_haplos[$j][0];
		}
		$marker=0; # this is needed in count_freqs for storage purposes
		$array_size=$#array_to_sort+1;
		count_freqs();
		sort_freqs();
		if($missing_check==1){
			print "0 removal (cases) for 1st marker  (option: miss 1):\n";
			print_freq();
		}
		# count freqs generates a variable called $TOT which is the total no of non-zero alleles
		for($j=0;$j<$no_case;$j++){
			if($trans_haplos[$j][0]==0){
				$new_allele=choose_allele();
				if($missing_check==1){
					print "Check: old allele = $trans_haplos[$j][0], new allele = $new_allele\n";
				}
				$trans_haplos[$j][0]=$new_allele;
			}
		}
		# all other markers
		for($i=1;$i<$mark;$i++){
			#@haplos_to_sort=();
			#for($j=0;$j<$no_case;$j++){
			#	$haplos_to_sort[$j][0]=$trans_haplos[$j][$i-1];
			#	$haplos_to_sort[$j][1]=$trans_haplos[$j][$i];
			#}
			#if($missing_check==1){
			#	print "haplos_to_sort:\n";
			#	for($j=0;$j<$no_case;$j++){
			#		print "($haplos_to_sort[$j][0],$haplos_to_sort[$j][1])\n";
			#	}
			#}
			#$haplo_array_size=$no_case;
			#$marker=$i; # this is needed in count_freqs for storage purposes
			#count_haplo_freqs();
			#if($missing_check==1){
			#	print "Before 0 removal (cases) (option: -miss 1):\n";
			#	print_haplo_freq();
			#}
			for($j=0;$j<$no_case;$j++){
				if($trans_haplos[$j][$i]==0){
					$k=0;
					@haplos_to_sort=();
					# searching through all the case haplotypes to find those that have the
					# same "haplotype" ie first allele.
					for($j1=0;$j1<$no_case;$j1++){
						if((($j1!= $j) && ($trans_haplos[$j1][$i-1]==$trans_haplos[$j][$i-1])) &&($trans_haplos[$j1][$i] !=0)){
							$haplos_to_sort[$k][0]=$trans_haplos[$j1][$i-1];
							$haplos_to_sort[$k][1]=$trans_haplos[$j1][$i];
							$k++;
						}
					}
					if($missing_check==1){
						print "haplos_to_sort:\n";
						for($l=0;$l<$k;$l++){
							print "haplo $l = ($haplos_to_sort[$l][0],$haplos_to_sort[$l][1])\n";
						}
					}
					$haplo_array_size=$k;
					$marker=$i; # this is needed in count_freqs for storage purposes
					# can choose one randomly..
					if($k>0){
						count_haplo_freqs();
						$new_allele=choose_haplo();
						if($missing_check==1){
							print "Before 0 removal (cases) - unique haplotype special case (option: -miss 1):\n";
							print "Problem missing haplotype is ($trans_haplos[$j][$i-1],$trans_haplos[$j][$i])\n";
							print "Change from trans_haplos[$j][$i]=$trans_haplos[$j][$i] to $new_allele\n";
						}
						$trans_haplos[$j][$i]=$new_allele;
					}
					else{	# need to resort to allele frequencies...
						@array_to_sort=();
						for($l=0;$l<$no_case;$l++){
							push @array_to_sort,$trans_haplos[$l][$i];
						}
						$array_size=$#array_to_sort+1;
						count_freqs();
						sort_freqs();
						if($missing_check==1){
							print "Allele distribution for unique haplotype\n";
							print_freq();
						}
						$new_allele=choose_allele();
						if($missing_check==1){
							print "Check: old allele = $trans_haplos[$j][$i], new allele = $new_allele\n";
						}
						$trans_haplos[$j][$i]=$new_allele;
					}
				}
			}
		}
		# ------------controls -------------
		# -------------special case of 1st marker
		@array_to_sort=();
		for($j=0;$j<$no_control;$j++){
			push @array_to_sort,$untrans_haplos[$j][0];
		}
		$array_size=$#array_to_sort+1;
		$marker=0; # this is needed in count_freqs for storage purposes
		count_freqs();
		sort_freqs();
		if($missing_check==1){
			print "0 removal for (controls) 1st marker (option: -miss 1):\n";
			print_freq();
		}
		# count freqs generates a variable called $TOT which is the total no of non-zero alleles
		for($j=0;$j<$no_control;$j++){
			if($untrans_haplos[$j][0]==0){
				$new_allele=choose_allele();
				if($missing_check==1){
					print "Check: old allele untrans_haplos[$j][0] = $untrans_haplos[$j][0], new allele = $new_allele\n";
				}
				$untrans_haplos[$j][0]=$new_allele;
			}
		}
		# --------- all other markers -----------
		for($i=1;$i<$mark;$i++){
			#@haplos_to_sort=();
			#for($j=0;$j<$no_control;$j++){
			#	$haplos_to_sort[$j][0]=$untrans_haplos[$j][$i-1];
			#	$haplos_to_sort[$j][1]=$untrans_haplos[$j][$i];
			#}
			#if($missing_check==1){
			#	print "haplos_to_sort:\n";
			#	for($j=0;$j<$no_control;$j++){
			#		print "($haplos_to_sort[$j][0],$haplos_to_sort[$j][1])\n";
			#	}
			#}
			#$marker=$i; # this is needed in count_freqs for storage purposes
			#$haplo_array_size=$no_control;
			#count_haplo_freqs();
			#if($missing_check==1){
			#	print "Before 0 removal (controls) (option: -miss 1):\n";
			#	print_haplo_freq();
			#}
			for($j=0;$j<$no_control;$j++){
				if($untrans_haplos[$j][$i]==0){
					$k=0;
					@haplos_to_sort=();
					# searching through all the case haplotypes to find those that have the
					# same "haplotype" ie first allele.
					for($j1=0;$j1<$no_control;$j1++){
						if((($j1!= $j) && ($untrans_haplos[$j1][$i-1]==$untrans_haplos[$j][$i-1])) && ($untrans_haplos[$j1][$i] != 0)){
							$haplos_to_sort[$k][0]=$untrans_haplos[$j1][$i-1];
							$haplos_to_sort[$k][1]=$untrans_haplos[$j1][$i];
							$k++;
						}
					}
					if($missing_check==1){
						print "haplos_to_sort:\n";
						for($l=0;$l<$k;$l++){
							print "haplo $l = ($haplos_to_sort[$l][0],$haplos_to_sort[$l][1])\n";
						}
					}
					$haplo_array_size=$k;
					$marker=$i; # this is needed in count_freqs for storage purposes
					# can choose one randomly..
					if($k>0){
						count_haplo_freqs();
						$new_allele=choose_haplo();
						if($missing_check==1){
							print "Before 0 removal (controls) - unique haplotype special case (option: -miss 1):\n";
							print "Unique haplotype is ($untrans_haplos[$j][$i-1],$untrans_haplos[$j][$i])\n";
							print "Change from untrans_haplos[$j][$i]=$untrans_haplos[$j][$i] to $new_allele\n";
						}
						$untrans_haplos[$j][$i]=$new_allele;
					}
					else{	# need to resort to allele frequencies...
						@array_to_sort=();
						for($l=0;$l<$no_control;$l++){
							push @array_to_sort,$untrans_haplos[$l][$i];
						}
						$array_size=$#array_to_sort+1;
						count_freqs();
						sort_freqs();
						if($missing_check==1){
							print "Allele distribution for unique haplotype\n";
							print_freq();
						}
						$new_allele=choose_allele();
						if($missing_check==1){
							print "Check: old allele = $untrans_haplos[$j][$i], new allele = $new_allele\n";
						}
						$untrans_haplos[$j][$i]=$new_allele;
					}
				}
			}
		}
	}
}

# sorts haplotypes, starting at marker $marker, and creates a series of vectors to describe the sort path
# the sort is per $marker and called in chi_square
sub sorter{
	my $i;
	my $j;
	my $k;
	my $l;
	my $x;
	my $y;
	my $t;
	my $oi;
	my $tmp;
	my $start_class=0;
	my $finish_class=0;
	my $start_class_controls=0;
	my $finish_class_controls=0;
	my $haplo_tmp;
	my $inds=0;
	my $inds_controls=0;
	my @inds_cnts=();
	my @inds_control_cnts=();
	my $sum=0;
	my @freqs_temp_controls=();
	my $array_size=0;
	my $array_size_controls=0;
	# initialising v and m arrays
	# m contains the ids of the haplotypes
	# v contains the class sizes
	# --------- initialisations ---------
	# for m[$marker][0][.] and v[$marker][0][.]	
	@array_to_sort=();
	@array_to_sort_controls=();
	for($j=0;$j<$no_case;$j++){
		$m[$marker][0][$j]=$j;
	}
	for($j=0;$j<$no_control;$j++){
		$m_controls[$marker][0][$j]=$j;
	}
	$v_controls[$marker][0][0]=$no_control;
	$v[$marker][0][0]=$no_case;
	# initialising store-haplotypes
	# only needs to be done once - for the original data (not any of the perms)
	if($sorter_check){
		$current_marker=0;
		print "Original dataset before sorting:\n";
		print_haplos_temp();
	}
	# --------FIRST MARKER SORT -----
	$oi=$order[$marker][0];
	$current_marker=$oi;
	# may need a special proviso here for the cases where there is only one class even after the first sort ie all alleles identical
	$start_class=0;
	$finish_class=$no_case;
	@array_to_sort=();
	for($l=0;$l<$v[$marker][0][0];$l++){
		$haplo_tmp=$m[$marker][0][$l+$start_class];
		push @array_to_sort,$trans_haplos[$haplo_tmp][$oi]; # note its $oi not $i as we need to sort according to a certain order
	}
	$array_size=$#array_to_sort+1;
	if($ah==0){ # start sort by most frequent
		count_freqs();
		sort_freqs();
	}
	if($ah==1){ # sort by given ancestral haplotype (fixed)
		count_freqs();
		sort_freqs_by_anchap();
	}
	if($ah==2){ # start sort by most significant allele in single marker test
		@case_freq_alls=();
		@control_freq_alls=();
		@case_unique_alleles=();
		@control_unique_alleles=();
		count_freqs();
		sort_freqs();
		@case_unique_alleles=@unique_alleles;
		@case_freq_alls=@freq_alls;
		@array_to_sort=();
		@unique_alleles=();
		$array_size=$#array_to_sort+1;
		$start_class_controls=0;
		$finish_class_controls=$no_control;
		for($l=0;$l<$v_controls[$marker][0][0];$l++){
			$haplo_tmp=$m_controls[$marker][0][$l+$start_class];
			push @array_to_sort,$untrans_haplos[$haplo_tmp][$oi]; # note its $oi not $i as we need to sort according to a certain order
		}
		count_freqs();
		sort_freqs();
		@control_unique_alleles=@unique_alleles;
		#print "case_unique_alleles=@case_unique_alleles\n";
		#print "control_unique_alleles=@control_unique_alleles\n";
		@control_freq_alls=@freq_alls;
		sort_freqs_by_chi();
		#print "case_freq_alls after sort_freqs_by_chi=@case_freq_alls\n";
		#print "control_freq_alls after sort_freqs_by_chi=@control_freq_alls\n";
		#exit(1);
		# must reassign array to sort to case data
		$start_class_controls=0;
		$finish_class_controls=$no_case;
		@array_to_sort=();
		for($l=0;$l<$v[$marker][0][0];$l++){
			$haplo_tmp=$m[$marker][0][$l+$start_class];
			push @array_to_sort,$trans_haplos[$haplo_tmp][$oi]; # note its $oi not $i as we need to sort according to a certain order
		}
	} 
	$inds=0;
	# internal check: The first marker to sort on must have at least as
	# many polymorphisms as number of classes, if not kick out
	if($c>($#unique_alleles+1)){
		print "Problem at marker $marker: Not polymorphic enough (number of different alleles at this marker = ",$#unique_alleles+1;
		print ") to allow $c classes.\n";
		exit(1);
	}
	# classes are determined by taking initially the c most frequent alleles at the first marker
	# ------------SORTING THE CASE HAPLOS FOR THE FIRST MARKER----------------
	# this is a special sort since we determine which alleles will determine the initial
	# clusters.
	# this sort reorders the m vector as appropriate
	# storing the original ancestral haplotypes 
	if($orig_sample){
		for($j=0;$j<$c;$j++){ # sort on only c+1 classes
			$store_haplotypes[$marker][$oi][$j]=$unique_alleles[$j];
		}
	}
	# note I've put the array loop and the cluster loop the other way around here to the main
	# sort text.
	# some initialisations
	$inds_cnts[0]=0;
	$sum=0;
	for($j=0;$j<$c;$j++){
		$sum+=$freq_alls[$j];
		$inds_cnts[$j+1]=$sum;
		$v[$marker][1][$j]=0;
		$v_controls[$marker][1][$j]=0;
	}
	$v[$marker][1][$c]=0;
	$v_controls[$marker][1][$c]=0;
	IND: for($x=0;$x<=$#array_to_sort;$x++){
		for($j=0;$j<$c;$j++){ # sort on only c+1 classes
			if($array_to_sort[$x]==$unique_alleles[$j]){
				# re-sort the relevant part of the m vector and put into the new m vector at sort marker $i+1
				$m[$marker][1][$inds_cnts[$j]]=$m[$marker][0][$x];
				$v[$marker][1][$j]+=1; # one more to that v class				
				$inds_cnts[$j]+=1;
				next IND;
			}
		}
		# the remaining haplotypes are allocated to the end of the m array or the cth class for the v array
		$m[$marker][1][$no_case-$v[$marker][1][$c]-1]=$m[$marker][0][$x];
		$v[$marker][1][$c]+=1;
	}
	# internal check:
	#----------SORTING THE CONTROL HAPLOS FOR THE FIRST MARKER-----------------------
	$start_class_controls=0;
	$finish_class_controls=$no_control;
	for($l=0;$l<$v_controls[$marker][0][0];$l++){
		$haplo_tmp=$m_controls[$marker][0][$l+$start_class];
		push @array_to_sort_controls,$untrans_haplos[$haplo_tmp][$oi]; # note its $oi not $i as we need to sort according to a certain order
	}
	$array_size_controls=$#array_to_sort_controls+1;
	$inds_control_cnts[0]=0;
	$sum=0;
	@freqs_temp_controls=();
	for($j=0;$j<$c;$j++){
		$freqs_temp_controls[$j]=0;
	}
	IND:for($x=0;$x<$array_size_controls;$x++){
		for($j=0;$j<$c;$j++){
			if($array_to_sort_controls[$x]==$unique_alleles[$j]){
				$freqs_temp_controls[$j]+=1;
				next IND;
			}
		}
		$freqs_temp_controls[$c]+=1;
	}
	for($j=0;$j<$c;$j++){
		$sum+=$freqs_temp_controls[$j];
		$inds_control_cnts[$j+1]=$sum;
	}
	IND_CONTROL: for($x=0;$x<$array_size_controls;$x++){
		for($j=0;$j<$c;$j++){ # sort on only c+1 classes
			if($array_to_sort_controls[$x]==$unique_alleles[$j]){
				# re-sort the relevant part of the m vector and put into the new m control vector at sort marker $i+1
				$m_controls[$marker][1][$inds_control_cnts[$j]]=$m_controls[$marker][0][$x];
				$v_controls[$marker][1][$j]+=1;
				$inds_control_cnts[$j]+=1;
				next IND_CONTROL;	
			}
		}
		$m_controls[$marker][1][$no_control-$v_controls[$marker][1][$c]-1]=$m_controls[$marker][0][$x];
		$v_controls[$marker][1][$c]+=1;
	}
	# checks:
	if($sorter_check){
		print "m[$marker][1] vector after sort\n";
		for($t=0;$t<$no_case;$t++){
			print "$m[$marker][1][$t] ";
		}
		print "\n";    
		print "v[$marker][1] vector after sort\n";
		for($t=0;$t<$classes;$t++){
			print "$v[$marker][1][$t] ";
		}
		print "\n";
		$current_marker=1;
		print "m_controls[$marker][1] vector after sort\n";
		for($t=0;$t<$no_control;$t++){
			print "$m_controls[$marker][1][$t] ";
		}
		print "\n";    
		print "v_controls[$marker][1] vector after sort\n";
		for($t=0;$t<$classes;$t++){
			print "$v_controls[$marker][1][$t] ";
		}
		print "\n";
		$current_marker=1;
		print "Case and control haplotypes after sorting on marker 1\n";
		print_haplos_temp();
	}
	# checking to see if we should continue sorting, this is only relevant
	# if any of the classes are not singletons.
	# changed to sharing 5 % or more ...
	$tmp=0;
	CLUSTER:for($j=0;$j<$classes;$j++){
		#print "v[$marker][1][$j]=$v[$marker][1][$j],share=$share\n";
		if($v[$marker][1][$j]>=$share){
			last CLUSTER;  #Still sharing, continue immediately
		}
		else{
			$tmp+=1;
		}
	}
	if($tmp==$classes){ # none are sharing anymore, quit
		$stop_sharing_marker[$marker]=1;
		if($orig_sample){
			$stop_sharing_marker_orig[$marker]=1;
			for($j=0;$j<$c;$j++){ 
				# undefine haplotype since not enough are sharing it anymore!
				delete($store_haplotypes[$marker][$oi][$j]);
			}
		}
		#print "Sharing has stopped for marker $marker at marker 1.\n";
		return; # only returns (i.e. doesn't sort further, if all of the above have failed.
	}
	# ---------SORTING ACCORDING TO ALL THE OTHER MARKERS-----------------------
	# now doing the sorting according to all the other markers ie not the central marker.
	# Here we now have a maximum of c+1=classes classes
	MARKER: for($i=1;$i<$mark;$i++){ # sorting, according to marker order, starting with the central marker
		$oi=$order[$marker][$i];
		$current_marker=$oi;
		$start_class=0;
		$finish_class=0;
		$start_class_controls=0;
		$finish_class_controls=0;
		$inds=0;
		$inds_controls=0;
		# may need a special proviso here for the cases where there is only one class even after the first sort ie all alleles identical
		# arranging the OTHER category ie class c:
		# this translates to a top down approach- whatever is already in c
		# stays in c and we can add to v and m above these haplotypes by then starting
		# bottom up
		# for cases
		for($l=0;$l<$v[$marker][$i][$c];$l++){
			$m[$marker][$i+1][$no-1-$l]=$m[$marker][$i][$no-1-$l];
		}
		$v[$marker][$i+1][$c]=$v[$marker][$i][$c];
		# for controls
		for($l=0;$l<$v_controls[$marker][$i][$c];$l++){
			$m_controls[$marker][$i+1][$no_control-1-$l]=$m_controls[$marker][$i][$no_control-1-$l];
		}
		$v_controls[$marker][$i+1][$c]=$v_controls[$marker][$i][$c];
		CLASS: for($k=0;$k<$c;$k++){
			$v[$marker][$i+1][$k]=0;
			$v_controls[$marker][$i+1][$k]=0;
			$start_class=$finish_class;
			$finish_class=$start_class+$v[$marker][$i][$k];
			$start_class_controls=$finish_class_controls;
			$finish_class_controls=$start_class_controls+$v_controls[$marker][$i][$k];
			if($sorter_check){
				print "(start_class,finish_class) for class $k=($start_class,$finish_class)\n";
				print "(start_class_controls,finish_class_controls) for class $k=($start_class_controls,$finish_class_controls)\n";
			}
			@array_to_sort=();
			for($l=0;$l<$v[$marker][$i][$k];$l++){
				$haplo_tmp=$m[$marker][$i][$l+$start_class];
				push @array_to_sort,$trans_haplos[$haplo_tmp][$oi]; # note its $oi not $i as we need to sort according to a certain order
			}
			$array_size=$#array_to_sort+1;
			count_freqs();
			if($ah==0||$ah==2){ # start sort by most frequent
				sort_freqs();
			}
			if($ah==1){ # sort by given ancestral haplotype (fixed)
				sort_freqs_by_anchap();
			}
			#if($ah==2){ # start sort by most significant allele in single marker test
			#	sort_freqs_by_chi();
			#} 
			if($sorter_check){
				print_freq();
			}
			if($orig_sample){
				$store_haplotypes[$marker][$oi][$k]=$unique_alleles[0];
				#print " for marker $i store_haplotypes[$marker][$oi][$k]=$store_haplotypes[$marker][$oi][$k]\n";
			}
			# ----------------------the ACTUAL sorting!!!!--------------------------------------------------
			# only sort to the first major allele after the first sort since |v_i|=c+1 from now on.
			# any haplotypes that don't match get moved to the OTHER allele category ie category c+1
			# this first sort reorders the m vector as appropriate
			for($x=0;$x<$array_size;$x++){
				# do not change unique_alleles! after the initial sort it is always the first
				# allele within the subsort that we are after...
				if($array_to_sort[$x]==$unique_alleles[0]){
					# re-sort the relevant part of the m vector and put into the new m vector at sort marker $i+1
					$m[$marker][$i+1][$inds]=$m[$marker][$i][$x+$start_class];
					$v[$marker][$i+1][$k]+=1; # one more to that v class
					$inds+=1;
				}
				else{ # this haplotype is now lost as it doesn't share anymore and cannot go to any  of the other classes
					$m[$marker][$i+1][$no_case-$v[$marker][$i+1][$c]-1]=$m[$marker][$i][$x+$start_class];
					$v[$marker][$i+1][$c]+=1;
				}
			}
			# the control haplotypes in the main sort part:
			if($v_controls[$marker][$i][$k]>1){
				@array_to_sort_controls=();
				#print "control haplos to sort on:\n";
				for($l=0;$l<$v_controls[$marker][$i][$k];$l++){
					$haplo_tmp=$m_controls[$marker][$i][$l+$start_class_controls];
					push @array_to_sort_controls,$untrans_haplos[$haplo_tmp][$oi]; # note its $oi not $i as we need to sort according to a certain order
				}
				$array_size_controls=$#array_to_sort_controls+1;
				for($x=0;$x<$array_size_controls;$x++){
					if($array_to_sort_controls[$x]==$unique_alleles[0]){
						# re-sort the relevant part of the m vector and put into the new m control vector at sort marker $i+1
						$m_controls[$marker][$i+1][$inds_controls]=$m_controls[$marker][$i][$x+$start_class_controls];	
						$v_controls[$marker][$i+1][$k]+=1;
						$inds_controls+=1;
					}
					else{ # this haplotype is now lost as it doesn't share anymore.
						$m_controls[$marker][$i+1][$no_control-$v_controls[$marker][$i+1][$c]-1]=$m_controls[$marker][$i][$x+$start_class_controls];
						$v_controls[$marker][$i+1][$c]+=1;
					}
				}
			}
			else{ # don't sort controls if there is only one or none left
				$v_controls[$marker][$i+1][$k]=$v_controls[$marker][$i][$k];	
				if($v_controls[$marker][$i][$k]>0){ # special case when only one (don't sort - just test)
					$haplo_tmp=$m_controls[$marker][$i][$start_class_controls];
					if($untrans_haplos[$haplo_tmp][$oi]==$unique_alleles[0]){
						$m_controls[$marker][$i+1][$inds_controls]=$m_controls[$marker][$i][$start_class_controls];
						$inds_controls+=1;
					}
					else{
						$m_controls[$marker][$i+1][$no_control-$v_controls[$marker][$i+1][$c]-1]=$m_controls[$marker][$i][$start_class_controls];
						# if the last haplotype is lost from this cluster it moves to the 
						# "other" cluster
						$v_controls[$marker][$i+1][$k]=0; 
						$v_controls[$marker][$i+1][$c]+=1;
					}
				}
			}
		}
		# print checks of updated m's and v's post sorting
		if($sorter_check){
			print "m[$marker][$i+1] vector after sort on marker $i+1\n";
			for($t=0;$t<$no;$t++){
				print "$m[$marker][$i+1][$t] ";
			}
			print "\n";    
			print "v[$marker][$i+1] vector after sort on marker $i+1\n";
			for($t=0;$t<$classes;$t++){
				print "$v[$marker][$i+1][$t] ";
			}
			print "\n";
			print "m_controls[$marker][$i+1] vector after sort on marker $i+1\n";
			for($t=0;$t<$no_control;$t++){
				print "$m_controls[$marker][$i+1][$t] ";
			}
			print "\n";    
			print "v_controls[$marker][$i+1] vector after sort on marker $i+1\n";
			for($t=0;$t<$classes;$t++){
				print "$v_controls[$marker][$i+1][$t] ";
			}
			print "\n";
			$current_marker=$i+1;
			print "Case and control haplotypes after sorting on marker $i+1\n";
			print_haplos_temp();
		}
		# test to see that we have reached the limit in ordering according to share_total 
		# quit sorting if this is the case.
		$tmp=0;
		CLUSTER1:for($j=0;$j<$c;$j++){
			if($v[$marker][$i+1][$j]>=$share){
				last CLUSTER1; # check where to quit...
			}
			else{
				$tmp+=1;
			}
		}
		if($tmp==$c){ # none are sharing anymore, quit
			# may have to change this to reflect the fact that we
			# only one delimiter for all classes for the sharing.
			$stop_sharing_marker[$marker]=$i+1;
			if($orig_sample){
				$stop_sharing_marker_orig[$marker]=$i+1;
				for($j=0;$j<$c;$j++){ 
					# undefine haplotype since not enough are sharing it anymore!
					delete($store_haplotypes[$marker][$oi][$j]);
				}
			}
			#print "Sharing has stopped for marker $marker at marker $i+1\n";
			return; # only returns if all of the above have failed.
		}
		# need to do this when the sharing after the last sort still exceeds 
		# sharing threshhold
		$stop_sharing_marker[$marker]=$mark;
		if($orig_sample){
			$stop_sharing_marker_orig[$marker]=$mark;
		}	
	}
}



# chi-square test for allele x versus all non x alleles
# used in all searches past first marker for all ah options
sub chi_square_stat{
my $l;
my $k;
my $test_stat=0;
my $N;
	$N=$obs[0][0]+$obs[0][1]+$obs[1][0]+$obs[1][1];
	if($stat_check==1){
		print "Observed values for cluster c statistic:\n";
		print "$obs[0][0]\t$obs[0][1]\n";
		print "$obs[1][0]\t$obs[1][1]\n";
		print "total sample size = $N\n";
	}
	# expected values for cluster i
	$exp[0][0]=($obs[0][0]+$obs[0][1])*($obs[0][0]+$obs[1][0])/$N; # A*C/N
	$exp[0][1]=($obs[0][0]+$obs[0][1])*($obs[0][1]+$obs[1][1])/$N; # A*D/N
	$exp[1][0]=($obs[1][0]+$obs[1][1])*($obs[0][0]+$obs[1][0])/$N; # B*C/N
	$exp[1][1]=($obs[1][0]+$obs[1][1])*($obs[0][1]+$obs[1][1])/$N; # B*D/N
	if($stat_check==1){
		print "Expected values for cluster c statistic:\n";
		print "$exp[0][0]\t$exp[0][1]\n";
		print "$exp[1][0]\t$exp[1][1]\n";
	}
	for($l=0;$l<2;$l++){
		for($k=0;$k<2;$k++){
			$test_stat+=($obs[$l][$k]-$exp[$l][$k])**2/$exp[$l][$k];
		}
	}
	return($test_stat);
}

# ---------------permutation tests --------------------------------------

# the test statistic.....calculated once for each marker
sub chi_square_test{
	my $i;
	my $j;
	my $k;
	my $l;
	my @chi_square=();
	my $N;
	$chi_square_sum=0;
	# marker is the current marker to sort on
	if($first_check){
		print "Marker ",$marker+1," $marker_names[$marker]\n";
	}
	sorter();
#	if($sorter_check){
#		print_sorter();
#	}
	# collapsing the likelihood over each allele for a one df per allele test for $c classes
	# don't do the first one ie j=1 as this is the initial position.
	# it results in a strange matrix where all cases and controls have the "same"
	# haplotype (because we currently have no information on which to sort them)
	#print "stop_sharing_marker[$marker]=$stop_sharing_marker[$marker]\n";
	MARKER_CHI: for($j=1;$j<$stop_sharing_marker[$marker];$j++){
		# observed counts
		for($i=0;$i<$c;$i++){# one 2 by 2 chi square table for each cluster
			$chi_square[$j-1][$i]=0;
			# observed values for cluster i
			$obs[0][0]=$v[$marker][$j][$i];
			$obs[0][1]=$no-$v[$marker][$j][$i];
			$obs[1][0]=$v_controls[$marker][$j][$i];
			$obs[1][1]=$no_control-$v_controls[$marker][$j][$i];
			if(($obs[1][1]==0) && ($obs[0][1]==0)){
				print "Problem with this marker: Not polymorphic between the case and control samples\n";
				print "Please edit your data to exclude this marker.\n";
				exit(1);
			}
			$N=$no_control+$no;
			$chi_square[$j-1][$i]=chi_square_stat();
			if($orig_sample){
				$chi_square_orig_parts[$marker][$j-1][$i]=$chi_square[$j-1][$i];
			}
			$chi_square_sum+=$chi_square[$j-1][$i];
		}
		# test to see that we have reached the limit in ordering, quit sorting if this is the case.
		for($i=0;$i<$c;$i++){
			if($v[$marker][$j][$i]>1){
				#print "continuing within chi-square\n";
				next MARKER_CHI;
			}
		}
		last; # only returns if all of the above have failed.
	}	
}

# permutation test for summary statistic.
# currently only permutes the entire sample once, rather than per statistic which leads to p-values which will be slightly correlated, but asymptoting
# to independence
sub perm_test{
	my $i;
	my $j;
	my $i1;
	my $j1;
	my $n1;
	my $n2;
	my $m1;
	my @total_array=();
	my $temp=0;
	my $total=$no_case+$no_control;
	my $new;
	my $counter=0;
	my $testing;
	my $testSD=0;
	# making a new set of case and control haplotypes
	# initialisations
	for($j1=0;$j1<$mark;$j1++){	
		$test_marker[$j1]="";
	}
	for($i=0;$i<$perms;$i++){
		$total=$no_case+$no_control;		
		if(($i+1)% 1000 ==0 ||$stat_check==1){
			print "permutation # ",$i+1,"\n";
		}
		# test to see if we need continue with permutation test: (another check
		# exists later on...
		if($i>1000){
			$testing=0;
			MARKER:for($j1=0;$j1<$mark;$j1++){
				if(!$test_marker[$j1]){  # continue
					$testing+=1;
					last MARKER;
				}
			}
			# only kick out if all perm tests are done
			if($testing == 0){
				print "No more permutation tests will be carried out as s.d. level of estimator has been achieved.\n";
				return; # might as well return
			}
		}
		# -------making the new data sets for the permutation-----------
		# resetting the joint sample contained in all_haplos
		for($i1=0;$i1<$total;$i1++){
			$total_array[$i1]=$i1;
		}
		# this finds new arrays by selecting cases (denoted by a 1) at random,
		# use controls as it means usually fewer random number calls.
		for($j=0;$j<$no_control;$j++){
			# int rounds down to the nearest integer
			$new=rand($total);
			#print "new=$new\n";
			# splice it out of the array:
			$temp=splice(@total_array,int($new),1);
			# temp is the ID no of haplotype chosen
			if($temp<$no_control){ # drawn from the control haplotypes
				for($j1=0;$j1<$mark;$j1++){
					$new_controls[$j][$j1]=$untrans_haplos_orig[$temp][$j1];
				}
			}
			else{ # drawn from the case haplotypes
				for($j1=0;$j1<$mark;$j1++){
					$new_controls[$j][$j1]=$trans_haplos_orig[$temp-$no_control][$j1];
				}
			}
			$total-=1; # one less to chose from now (so doing it without redraw)
		}
		# the leftovers become the casehaplos:
		#check:
		if(($#total_array+1) != $no_case){
			print "Problem in allocation of new perm set\n";
			print "No_cases = $no_case, haplos left to allocate = ",$#total_array+1;
			print "\n";
			exit(1);
		}
		for($j=0;$j<$no_case;$j++){
			if($total_array[$j]<$no_control){
				for($j1=0;$j1<$mark;$j1++){
					$new_cases[$j][$j1]=$untrans_haplos_orig[$total_array[$j]][$j1];
				}
			}
			else{
				for($j1=0;$j1<$mark;$j1++){
					$new_cases[$j][$j1]=$trans_haplos_orig[$total_array[$j]-$no_control][$j1];
				}
			}
		}
		# for error testing
		if($perm_check){
			print_new_haplos();
		}
		# ---------calculating the test statistic------------
		# question: What is correct ?
		# permuting once per sample and calculating all the markers sharing stats
		# or doing each one separately (will run much longer)
		# leaving as the former at the moment.
		# calculating the test statistic from each marker
		@trans_haplos=@new_cases;
		@untrans_haplos=@new_controls;
		$no=$no_case;
		#initialise();
		@v=();
		@v_controls=();
		for($marker=0;$marker<$mark;$marker++){
			$counter=0;
			if(!$test_marker[$marker] || $i<1000){ # do at least 1000 perms.
				chi_square_test();
				$chi_square_perms[$i][$marker]=$chi_square_sum;
				if($stat_check==1){
					print "chi-square test statistic chi_square_perms[$i][$marker] = $chi_square_perms[$i][$marker]\n";
				}
				if($stat_check==1){
					print "chi_square_perms[$i][$marker]=$chi_square_perms[$i][$marker]\n";
					print "chi_square_orig[$marker]=$chi_square_orig[$marker]\n";
				}
				if($chi_square_perms[$i][$marker]>=$chi_square_orig[$marker]){
					$counter=1;
				}
				else{
					$counter=0;
				}
				# take care here: we want sd and mean of estimator of phat a bernoulli var
				# and we are taking a recursive mean and sd . If sd is lower than a set value
				# do not do anymore perm tests on this marker but continue instead.
				# recursively calculated mean:
				if($i>0){
					$chi_square_perms_mean[$marker]=$counter/($i+1)+$i*$chi_square_perms_mean[$marker]/($i+1);
					$chi_square_perms_se[$marker]=sqrt($chi_square_perms_mean[$marker]*(1-$chi_square_perms_mean[$marker])/($i+1));
				}
				else{
					$chi_square_perms_mean[$marker]=$counter;
					$chi_square_perms_se[$marker]=0;
				}
				if($i==0){
					$testSD=0.001;
				}
				else{ # make the SD of the estimator flexible to suit pha
					# make the Standard deviation 1/100th of the mean as the threshold when calculations cease.
					$testSD=$SD*$chi_square_perms_mean[$marker];
				}
				if($stat_check){
					print "Perm = ",$i+1, "counter = $counter, Marker $marker, mean = ";
					printf "%1.2lf",$chi_square_perms_mean[$marker];
					print ",s.d. = ";
					printf "%1.2f",$chi_square_perms_se[$marker];
					print "\n"; 
				}
				# printing all the perm chi square to a file (original chi squares are the top line
				printf OUT2 "%1.2f",$chi_square_perms[$i][$marker];
				print OUT2 " ";
				if(($chi_square_perms_se[$marker]<$testSD && $i>=1000) && $chi_square_perms_mean[$marker]>0){
					print "Permutation tests ceased at permutation $i for marker ",$marker+1,".\n";
					$test_marker[$marker]=$i; # permutation test is over for this marker at perm $i
				}
			}
			else{ # no value calculated because sd of estimator already pretty low...
				print OUT2 "    ";
			}
		}
		print OUT2 "\n";
	}
}

# ------------------------------------the main subroutine--------------------------------------
sub main_perm{
	my $time2;
	my $m;
	my $i;
	my @local_time1;
	my $j;
	# ----------original test-----------------------------
	$no=$no_case;
	for($i=0;$i<$no_case;$i++){
		for($m=0;$m<$mark;$m++){
			$trans_haplos[$i][$m]=$trans_haplos_orig[$i][$m];
		}
	}
	for($i=0;$i<$no_control;$i++){
		for($m=0;$m<$mark;$m++){
			$untrans_haplos[$i][$m]=$untrans_haplos_orig[$i][$m];
		}
	}
	# print "Original dataset of haplotypes:\n";
	# set current marker to 0 to get before sorting.
	# print_orig_haplos();
	# inferring the missing data
	fix_missing_data(); 
	#print "Haplotype data after inferring data to replace missing marker information:\n";
	#print_orig_haplos();
	#exit(1);
	# I have to do this differently to before, I need a sort on the 
	# transmitted haplos and a sort on the untransmitted haplos,
	# the transmitted haplos will drive the sorting,ie determine the top clusters.
	print "Total number of haplotypes to which sharing is calculated = $share\n";
	$orig_sample=1; # can set this to any value to make it TRUE
	for($marker=0;$marker<$mark;$marker++){
		chi_square_test();
		$chi_square_orig[$marker]=$chi_square_sum;
		printf OUT2 "%1.2f",$chi_square_orig[$marker];
		print OUT2 " ";
		if($stat_check==1){
			print "Original Chi-square test result = $chi_square_sum, after sort emanating from marker ",$marker+1,"\n";
		}
	}
	print OUT2 "\n\n";
	# copying the original m,v's into m_orig, v_orig for safe keeping
	for($marker=0;$marker<$mark;$marker++){
		for($j=0;$j<=$stop_sharing_marker_orig[$marker];$j++){
			for($i=0;$i<$no_case;$i++){
				$m_orig[$marker][$j][$i]=$m[$marker][$j][$i];
			}
			for($i=0;$i<$no_control;$i++){
				$m_controls_orig[$marker][$j][$i]=$m_controls[$marker][$j][$i];
			}
			for($i=0;$i<=$c;$i++){
				$v_orig[$marker][$j][$i]=$v[$marker][$j][$i];
				$v_controls_orig[$marker][$j][$i]=$v_controls[$marker][$j][$i];
			}
		}
	}
	# reset the flag for haplotype storage - don't want to store perm haplotypes	
	$orig_sample=""; # reset to empty string -> false in PERL
	# ---------------perm test stuff----------------------
	if($perms<2){
		print "No permutation testing carried out as sample size is too small (n<2).\n";
		print "Original sample test statistics are evaluated.\n";
		results_print();
	}
	else{
		perm_test(); # generates the chi_square_perm array of simulated chi square values
		p_values_print();
	}
	$time2=time();
	print OUT1 "Time taken (m) = ";
	print "Time taken (m) = ";
	print OUT1 int(($time2-$time1)/60);
	print int(($time2-$time1)/60);
	print OUT1 ", ";
	print ", ";
	@local_time1=localtime($time2);
	print OUT1 "$local_time1[3]/",$local_time1[4]+1,"/",$local_time1[5]+1900,", ";
	print "$local_time1[3]/",$local_time1[4]+1,"/",$local_time1[5]+1900,", ";
	print OUT1 "$local_time1[2]:$local_time1[1]:$local_time1[0]\n";
	print "$local_time1[2]:$local_time1[1]:$local_time1[0]\n";
	print OUT1 "haplo_clusters.pl",qw($Revision: 1.27 $),"\n";
	print "haplo_clusters.pl",qw($Revision: 1.27 $),"\n";
}
